#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <errno.h>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "macros.h"
#include "common.h"
#include "complex.h"
#include "blt.h"
#include "iir_coef.h"
#include "filter_poles.h"
#include "landen_trans.h"

static void
frequency_response(const char *file, struct transfer_func *h, double freq_start, double freq_end, double step)
{
	complex_t s, p, numer, denom, h_omega;
	double magnitude, freq;
	double numphase, denphase;
	unsigned int i;
	FILE *fp;

	return_if_fail(h != NULL);
	return_if_fail(file != NULL); 

	fp = fopen(file, "w+");

	if (!fp)
		return;

	fprintf(fp, "# Frequency\tMagnitude\tPhase\n");
	
	for (freq = freq_start; freq <= freq_end; freq += step) {

		denom = c_num(1.0, 0.0);
		numer = c_num(h->h_zero, 0.0);
		s = c_num(0.0, freq);
		numphase = 0.0;
		denphase = 0.0;

		for (i = 0; i < h->nz; i++) {
			if (h->zeros[i].im != INFINITY) {
				p = c_sub(s, h->zeros[i]);
				numer = c_mult(numer, p);
				numphase += c_phase(p);
			}
		}

		for (i = 0; i < h->np; i++) {
			p = c_sub(s, h->poles[i]);
			denom = c_mult(denom, p);
			denphase += c_phase(p);
		}

		h_omega = c_div(numer, denom);
		magnitude = 20.0 * log10(c_magnitude(h_omega));
		//magnitude = c_magnitude(h_omega);

		fprintf(fp, "%3.2f\t\t%3.2f\t\t%3.2f\n", freq, magnitude, -1.0 * fabs(numphase-denphase) / M_PI * 180.0);
	}

	fclose(fp);
}

#define HAVE_FREQ_PASS1 	0x0001
#define HAVE_FREQ_STOP1		0x0002
#define HAVE_FREQ_PASS2 	0x0001
#define HAVE_FREQ_STOP2		0x0002
#define HAVE_ATTEN_STOP		0x0004
#define HAVE_RIPPLE		0x0008
#define HAVE_SAMPLING_RATE	0x0010
#define HAVE_FILTER_TYPE	0x0020
#define HAVE_APPROX_TYPE	0x0040
#define HAVE_USE_RORABAUGH_BLT	0x0080
#define HAVE_OUTPUT_RESPONSE	0x0100

#define MANDATORY_OPT_MASK	(HAVE_FREQ_PASS1 | HAVE_FREQ_STOP1 |HAVE_FREQ_PASS2 | HAVE_FREQ_STOP2 | HAVE_ATTEN_STOP | HAVE_SAMPLING_RATE)

char *program_name;

void
print_usage()
{
	printf("%s: -a As -p1 Fp1 -s1 Fs1 -p2 F2 -s2 Fs2 -o output -r ripple -t [hp|lp|bp|bs] -A [chebysh|butterw] -T sample -p\n", program_name);
	printf("\n");
	printf("Options:\n");
	printf(" -a As                  	  stopband attenuation, dB \n");
	printf(" -p Fp1                  	  low pass frequency, Hz \n");
	printf(" -s Fs1                  	  low stop frequency, Hz \n");
	printf(" -j Fp2                 	  high pass frequency, Hz \n");
	printf(" -k Fs2                  	  high stop frequency, Hz \n");
	printf(" -o output              	  write frequency response to file \n");
	printf(" -r ripple              	  ripple in passband, dB \n");
	printf(" -t [hp|lp|bp|bs]		  filter type hp = highpass, lp = lowpass, bp = bandpass, bs = bandstop \n");
	printf(" -A [chebysh|butterw|ellipt]	  approximation chebyshev, butterworth or elliptic\n");
	printf(" -T sample			  sampling rate, Hz \n");
	printf("\n");
	printf("Send bugs and reports to Andrey Turubanov, <parslein@yandex.ru>\n");
}

int
main(int argc, char *argv[])
{
	struct filter_spec *spec;
	struct transfer_func *h, *z;
	struct iir_filter_coef *iir;
	char *c, *freq_resp_file;
	double opt_freq_pass1, opt_freq_stop1, opt_freq_pass2, opt_freq_stop2;
	double opt_atten_stop, opt_ripple;
	double opt_sampling_rate;
	double sample_rate;
	double fp1, fp2, fs1, fs2;
	enum filter_type opt_type;
	enum filter_approx opt_approx;
	int i, opt, optmask;

	opterr = 0;
	optmask = 0;
	freq_resp_file = NULL;
	opt_freq_pass1 = opt_freq_stop1 = opt_freq_pass2 = opt_freq_stop2 = 0.0;
	opt_sampling_rate = opt_ripple = 0.0;
	opt_atten_stop = 0.0;
	opt_ripple = 0.1;
	opt_type   = LOWPASS;
	opt_approx = BUTTERWORTH;

	program_name = ((c = strrchr(argv[0], '/')) != NULL ? ++c: argv[0]);
	
	while ((opt = getopt(argc, argv, "a:p:s:j:k:o:r:t:A:T:1")) != -1) {
		switch (opt) {
		case 'a':
			opt_atten_stop = atof(optarg);
			optmask |= HAVE_ATTEN_STOP;
			break;
		case 'p':
			opt_freq_pass1 = atof(optarg);
			optmask |= HAVE_FREQ_PASS1;
			break;
		case 's':
			opt_freq_stop1 = atof(optarg);
			optmask |= HAVE_FREQ_STOP1;
			break;
		case 'j':
			opt_freq_pass2 = atof(optarg);
			optmask |= HAVE_FREQ_PASS2;
			break;
		case 'k':
			opt_freq_stop2 = atof(optarg);
			optmask |= HAVE_FREQ_STOP2;
			break;
		case 'o':
			freq_resp_file = optarg;
			optmask |= HAVE_OUTPUT_RESPONSE;
			break;
		case 'r':
			opt_ripple = atof(optarg);
			optmask |= HAVE_RIPPLE;
			break;
		case 't':
			if (strcmp(optarg, "lp") == 0)
				opt_type = LOWPASS;
			else if (strcmp(optarg, "hp") == 0)
				opt_type = HIGHPASS;
			else if (strcmp(optarg, "bp") == 0)
				opt_type = BANDPASS;
			else if (strcmp(optarg, "bs") == 0)
				opt_type = BANDSTOP;
			else {
				fprintf(stderr, "unknown filter type `%s'\n", optarg);
				print_usage();
				exit(1);
			}
			optmask |= HAVE_FILTER_TYPE;
			break;
		case 'A':
			if (strcmp(optarg, "chebysh") == 0)
				opt_approx = CHEBYSHEV;
			else if (strcmp(optarg, "butterw") == 0)
				opt_approx = BUTTERWORTH;
			else if (strcmp(optarg, "ellipt") == 0)
				opt_approx = ELLIPTIC;
			else {
				fprintf(stderr, "unknown approximation type `%s'\n", optarg);
				print_usage();
				exit(1);
			}
			optmask |= HAVE_APPROX_TYPE;
			break;
		case 'T':
			opt_sampling_rate = atof(optarg);
			optmask |= HAVE_SAMPLING_RATE;
			break;
		case '1':
			optmask |= HAVE_USE_RORABAUGH_BLT;
			break;
		default:
			print_usage();
			exit(1);
		}
	}

	/* Check options validity. */

	if ((optmask & MANDATORY_OPT_MASK) == MANDATORY_OPT_MASK) {

		if ((opt_type == BANDPASS) || (opt_type == BANDSTOP))
			if ((opt_freq_pass1 > opt_freq_pass2) || (opt_freq_stop1 > opt_freq_stop2))
				goto error;

		if ((opt_freq_pass1 <= 0.0) || (opt_freq_stop1 <= 0.0) || (opt_atten_stop <= 0.0) || (opt_sampling_rate <= 0.0))
			goto error;

		if ((opt_freq_pass2 >= opt_sampling_rate / 2.0) || (opt_freq_stop2 >= opt_sampling_rate / 2.0))
			goto error;

		switch (opt_type) {
		case LOWPASS:
			if (opt_freq_pass1 >= opt_freq_stop1)
				goto error;
			break;
		case HIGHPASS:
			if (opt_freq_stop1 >= opt_freq_pass1)
				goto error;
			break;
		case BANDPASS:
			if ((opt_freq_stop1 >= opt_freq_pass1) || (opt_freq_pass2 >= opt_freq_stop2))
				goto error;
			break;
		case BANDSTOP:
			if ((opt_freq_pass1 >= opt_freq_stop1) || (opt_freq_stop2 >= opt_freq_pass2))
				goto error;
			break;	
		}

		if ((optmask & HAVE_RIPPLE) && (opt_approx == CHEBYSHEV) && ((opt_ripple > 3.0) || (opt_ripple <= 0.0)))
			goto error;

	} else {
		error:
		fprintf(stderr, "filter parameters inconsistent\n");
		print_usage(1);
		exit(1);
	}

	/* Generate filter specification. */

	sample_rate = 1.0 / opt_sampling_rate;
	fp1 = 2.0 * M_PI * opt_freq_pass1;
	fs1 = 2.0 * M_PI * opt_freq_stop1;
	fp2 = 2.0 * M_PI * opt_freq_pass2;
	fs2 = 2.0 * M_PI * opt_freq_stop2;

	fp1 = bilinear_frequency_digital(fp1, sample_rate);
	fs1 = bilinear_frequency_digital(fs1, sample_rate);;
	fp2 = bilinear_frequency_digital(fp2, sample_rate);;
	fs2 = bilinear_frequency_digital(fs2, sample_rate);;

	spec = filter_make_spec(opt_type, opt_approx, opt_ripple, opt_atten_stop, fp1, fs1, fp2, fs2);
	
	if (spec == NULL) {
		fprintf(stderr, "%s: can't create filter specification\n", program_name);
		exit(1);
	}

	/* Obtain S-plane transfer function. */

	h = filter_analog(spec);
	
	if (h == NULL) {
		fprintf(stderr, "%s: can't create s-plane transfer function\n", program_name);
		exit(1);
	}

	if (optmask & HAVE_OUTPUT_RESPONSE)
		frequency_response(freq_resp_file, h, 0.01, M_PI, 0.01);

	/* Map S-plane to Z-plane. */

	if (optmask & HAVE_USE_RORABAUGH_BLT)
		z = bilinear_transform(h, sample_rate, TRUE);
	else
		z = bilinear_transform_v2(h, TRUE);

	if (z == NULL) {
		fprintf(stderr, "%s: can't map s-plane to z-plane\n", program_name);
		exit(1);
	}

	/* Calculate coefficients of IIR biquadratic cascades. */

	iir = iir_calc_coef(z);

	if (iir == NULL) {
		fprintf(stderr, "%s: can't calculate IIR coefficients\n", program_name);
		exit(1);
	}

	printf("        Filter parameters\n");
	printf("\n");
	printf("                     Type: %s\n", (opt_type == LOWPASS ? "lowpass" : (opt_type == HIGHPASS ? "highpass" : (opt_type == BANDPASS ? "bandpass" : (opt_type == BANDSTOP ? "bandstop" : "unknown")))));
	printf("            Approximation:  %s\n", (opt_approx == BUTTERWORTH ? "BUTTERW" : (opt_approx == CHEBYSHEV ? "CHEBYSH" : (opt_approx == ELLIPTIC ? "ELLIPT" : "unknown"))));
	printf("    Pass frequency (Hz): %8.2f\n", opt_freq_pass1);
	printf("      Stop frequency (Hz): %8.2f\n", opt_freq_stop1);
	printf("    High pass frequency (Hz): %8.2f\n", opt_freq_pass2);
	printf("    High stop frequency (Hz): %8.2f\n", opt_freq_stop2);
	printf("Stopband attenuation (dB): %8.2f\n", opt_atten_stop);
	
	if (opt_approx == CHEBYSHEV)
		printf("              Ripple (dB):      %2.1f\n", opt_ripple);
	if (opt_approx == ELLIPTIC)
		printf("              Ripple (dB):      %2.1f\n", opt_ripple);

	printf("       Sampling rate (Hz): %8.2f\n", opt_sampling_rate);
	printf("          Prototype order: %u\n", spec->order);
	printf("\n");

	if (h != NULL) {
		printf("        S-plane filter prototype H(s)\n");
		printf("\n");
		printf("  H0 = % 1.7f\n", h->h_zero);
		printf("\n");
		printf("         Poles                     Zeros\n");

		for (i = 0; i < h->np; i++) {
			if (i < h->nz) {
				printf(" % 1.7f % 1.7fi    % 1.7f % 1.7fi\n",
				       h->poles[i].re, h->poles[i].im, h->zeros[i].re, h->zeros[i].im);
			} else {
				printf(" % 1.7f % 1.7fi\n", h->poles[i].re, h->poles[i].im);
			}
		}

		printf("\n");
	}

	if (z != NULL) {
		printf("        Z-plane transfer function H(z)\n");
		printf("\n");
		printf("  H0 = % 1.7f\n", z->h_zero);
		printf("\n");
		printf("         Poles                     Zeros\n");

		for (i = 0; i < z->np; i++) {
			if (i < z->nz) {
				printf(" % 1.7f % 1.7fi    % 1.7f % 1.7fi\n",
				       z->poles[i].re, z->poles[i].im, z->zeros[i].re, z->zeros[i].im);
			} else {
				printf(" % 1.7f % 1.7fi\n", z->poles[i].re, z->poles[i].im);
			}
		}

		printf("\n");
	}

	if (iir != NULL) {
		printf("        IIR biquadratic 2-nd order cascades\n");
		printf("\n");

		for (i = 0; i < iir->nrcasc; i++) {
			printf("Stage %i \n", i+1);
			printf("\tb0 =% 1.7f\n", iir->casc[i].b0);
			printf("\tb1 =% 1.7f\ta1 =% 1.7f\n", iir->casc[i].b1, iir->casc[i].a1);
			printf("\tb2 =% 1.7f\ta2 =% 1.7f\n", iir->casc[i].b2, iir->casc[i].a2);
		}

		printf("\n");
	}

	if (spec != NULL)
		FREE(spec);

	if (h != NULL) {
		if (h->poles != NULL)
			FREE(h->poles);
		if (h->zeros != NULL)
			FREE(h->zeros);
		FREE(h);
	}

	if (z != NULL) {
		if (z->poles != NULL)
			FREE(z->poles);
		if (z->zeros != NULL)
			FREE(z->zeros);
		FREE(z);
	}

	if (iir != NULL) {
		if (iir->casc != NULL)
			FREE(iir->casc);
		FREE(iir);
	}

	return 0;
}
